/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	????
**
**FILE NAME:
**	CnxLoad.c
**
**ABSTRACT:
**	This file implements the download application for the Tigris ADSL
**	device.  It will open and read the hex record file, process the
**	record into binary data and then send them to the driver for
**	execution.
**
**DETAILS:
**	????
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Yukon-AV/Linux_2.4_ATM_Driver/DownLoadApp/cnxadslload.c $
**	$Revision: 4 $
**	$Date: 7/05/01 9:50p $
*******************************************************************************
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <getopt.h>
#include <wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <atm.h>

#include "../LnxDefs.h"
#include "../CommonData.h"
#include "../ConfigLabels.h"

#include "readuserparams.h"

#include "../hexformat.h"

// local functions

static DWORD
atoh(
	const char *	cInStr,
	DWORD		uNumChars);

static NTSTATUS
SendIoctlCmd(
	IN int			SocketID,
	IN int			CommandID,
	struct atmif_sioc	*pSocketioc);

static NTSTATUS
downLoadMicroCode(
	IN int		SocketID,
	IN int		FileHandle);

static NTSTATUS
SendUserParams(
	IN int				SocketID,
	IN const char *			Path,
	OUT unsigned long *		pAutoLogEnable,
	OUT char *			AutoLogPath,
	OUT char *			FirmwareFilename,
	OUT unsigned long *		pAutoLogFrequency,
	OUT unsigned long *		pAutoLogMaxSize,
	OUT unsigned long *		pAutoLogOverWrite);

static struct atmif_sioc		Socketioc;
static int				d_opt;

/******************************************************************************
FUNCTION NAME:
	main

ABSTRACT:
	This is main entrypoint for the download executable.  It will
	open the socket and make calls to first start the download,
	execute the download, and then complete the download.

PARAMETERS:
	argc - command line parameter count
	argv - command line parameters

RETURN:
	The status of the operation.

DETAILS:
******************************************************************************/
int
main(
	int		argc,
	char **		argv)
{
  NTSTATUS			NTStatus=STATUS_SUCCESS;
  int				SocketID;
  int				FileHandle;
  char				Dummy [8];
  char				FileName [512];
  char				FileNameFromConf [512];
  char				AutoLogFile [512];
  unsigned long			AutoLogEnable;
  unsigned long			AutoLogFrequency;
  unsigned long			AutoLogMaxSize;
  unsigned long			AutoLogOverwrite;

  int				c;
  int				h_opt = 0;
  int				opt_err = 0;

  while ((c = getopt(argc,argv,"dhi:")) != EOF) {
    switch (c) {
    case 'd' :
      d_opt |= 1;
      break;
    case 'h' :
      h_opt |= 1;
      opt_err |= 1;
      break;
    case 'i' :
      Socketioc.number = atoi(optarg);
      break;
    default :
      opt_err |= 1;
      break;
    }
  }

  if (opt_err || ((argc - optind) != 1)) {
    printf("Usage: cnxadslload [-d][-h][-i itf] Path\n");
    if (h_opt)
      printf("-d     debugging\n"
	     "-h     help\n"
	     "-i itf atm interface number\n");
    exit(1);
  }

  // open the socket to the device.  This socket
  // will be used to issue the load commands
  SocketID = socket(PF_ATMPVC, SOCK_DGRAM, 0);
  if (SocketID < 0)
  {
    printf("Socket open failure\n");
    return STATUS_FAILURE;
  }

  if (d_opt) printf("About to SendUserParams\n");
  // send the user configurable parameters to the
  // driver and get the autologging parameters
  NTStatus = SendUserParams(SocketID,
			    argv[optind],
			    &AutoLogEnable,
			    AutoLogFile,
			    FileNameFromConf,
			    &AutoLogFrequency,
			    &AutoLogMaxSize,
			    &AutoLogOverwrite);
  if (!NT_SUCCESS(NTStatus))
  {
    printf("send user parameters (SendUserParams) failed, NTStatus=%lx\n", NTStatus);
    return STATUS_FAILURE;
  }

  // open the firmware file and validate that we got it open
  strcpy(FileName, argv[optind]);
  strcat(FileName, "/");
  strcat(FileName, FileNameFromConf);
  FileHandle = open(FileName, O_RDONLY);
  if (FileHandle < 0)
  {
    printf("Firmware file not found (%s)\n", FileName);
    return STATUS_FAILURE;
  }

  // issue the prepare for download command
  Socketioc.length = sizeof (Dummy);	// so I am sending fake stuff
  Socketioc.arg = Dummy;  		// Linux doesn't like NULL here

  NTStatus = SendIoctlCmd(SocketID, TIG_LOAD_START, &Socketioc);
  // if status is success then it needs a download.
  // failure indicates that the download is already
  // done and therefore not required
  if (NT_SUCCESS(NTStatus))
  {
    if (d_opt) printf("About to downLoadMicroCode\n");

    // call a function to execute the download
    NTStatus = downLoadMicroCode(SocketID, FileHandle);

    // close the firmware file
    close(FileHandle);

    if (!NT_SUCCESS(NTStatus))
    {
      printf(" download ARM micro code (downLoadMicroCode) failed, NTStatus=%lx\n", NTStatus);
      return STATUS_FAILURE;
    }

    // issue the download complete command
    NTStatus = SendIoctlCmd(SocketID, TIG_LOAD_DONE, &Socketioc);
    if (!NT_SUCCESS(NTStatus))
      return STATUS_FAILURE;
  }
  else
  {
    // close the firmware file
    close(FileHandle);

    printf(" download is already done\n");
  }

  // issue the start dsl function command
  NTStatus = SendIoctlCmd(SocketID, TIG_START_DSL, &Socketioc);
  if (!NT_SUCCESS(NTStatus))
    return STATUS_FAILURE;
  else
    printf("\nConexant AccessRunner load complete!\n");

  // close out this socket
  close(SocketID);

  return 0;
}

/****************************************************************************
The following functions	were taken directly from the NB	RipTide
Cmdif.c	/ Cmdif.h files then modified to allow either CR/LF or just LF for 
Linux
****************************************************************************/

/******************************************************************************
FUNCTION NAME:
	downLoadMicroCode

ABSTRACT:
	Use IFSMgr services to download	a controller
	code HEX file to the RAM of the device

PARAMETERS:
	char *sFN is the pathname of the file

RETURN:
	Error code 0 --	success, otherwise fail

DETAILS:
******************************************************************************/
NTSTATUS
downLoadMicroCode(
	IN int		SocketID,
	IN int		FileHandle)
{
  DWORD			BytesRead;
  DWORD			dwSegAddr = 0;	// Extended segment address
  DWORD			dwLinAddr = 0;	// Extended linear address
  DWORD			uRecType;	// extracted record type
  DWORD			dwIPValue;
  BOOL			FoundLastRecord = FALSE;
  HEX_HEADER_T *       	pRecHeader;	// ptr to location in LoadRecord of hex header
  DWORD			RecordLen;	// extracted and coverted to binary record length
  NTSTATUS		NTStatus = STATUS_SUCCESS;
  CHAR *	       	pBuffer;	// ptr to buffer within LoadRecord
  TIG_LOAD_RECORD_T *	pLoadRecord = NULL;	// image of one line from hex file
  BOOL			MarkFound;

  //
  // Setup
  //

  // allocate a record and point the header into it
  pLoadRecord = malloc(MAX_RECORD_LENGTH);
  if (!pLoadRecord)
  {
    printf("Error - Insufficient memory, Load terminating\n");
    return -ENOMEM;
  }

  pRecHeader = (HEX_HEADER_T*) &pLoadRecord->RecordData[0];
  pBuffer = &pLoadRecord->RecordData[0];

  //
  //  Read file and process each line
  //
  do
  {
    // read char by char until we find Record Header Mark skipping all whitespace
    MarkFound = FALSE;
    while (TRUE)
    {
      //* Read 1 byte searching for Record Header mark
      BytesRead = read(FileHandle, pRecHeader->Mark, 1);
      if (BytesRead != 1)
      {
        NTStatus = STATUS_FAILURE;
        printf("Error - Read header fail, Load terminating\n");
        break;
      }

      // exit if Mark found
      if (*pRecHeader->Mark == ':')
      {
        MarkFound = TRUE;
        break;
      }

      // otherwise must be whitespace
      if (!isspace(*pRecHeader->Mark))
      {
        NTStatus = STATUS_FAILURE;
        printf("Error - Read header fail, Load terminating\n");
        break;
      }
    }
    if (!MarkFound)
      break;

    //* Terminate Buffer to ensure we find good data
    pRecHeader->Mark[1] = 0x00;

    //* Read Record Header into location after Mark
    BytesRead = read(FileHandle, &pRecHeader->Mark[1], sizeof(HEX_HEADER_T)-1);

    if (BytesRead != (sizeof(HEX_HEADER_T)-1))
    {
      NTStatus = STATUS_FAILURE;
      printf("Error - Read header fail, Load terminating\n");
      break;
    }

    // Determine variable record length if applicable
    RecordLen = atoh(pRecHeader->Length, REC_DATA_LEN);
    RecordLen *= 2;        // Two Bytes are used to represent each hex digit

    //* Read The Rest of the Record. At minimum, read Checksum
    RecordLen += REC_CHECKSUM_LEN;
    if (MAX_RECORD_LENGTH < (sizeof(HEX_HEADER_T) + sizeof(TIG_LOAD_RECORD_T) + RecordLen))
    {
      NTStatus = STATUS_FAILURE;
      printf("Error - Record length error, Load terminating\n");
      break;
    }

    BytesRead = read(FileHandle, pBuffer + sizeof(HEX_HEADER_T), RecordLen);

    if (BytesRead != RecordLen)
    {
      NTStatus = STATUS_FAILURE;
      printf("Error - Record read failure, Load terminating\n");
      break;
    }

    //* Do not consider Checksum as part of the real data record
    RecordLen -= REC_CHECKSUM_LEN;

    //* Parse the record type
    uRecType = atoh(pRecHeader->Type, REC_TYPE_LEN);

    switch(uRecType)
    {
      case DATA_REC:
        //*
        //* Data Record
        //*
        pLoadRecord->RecordType  	= DATA_REC;
      	pLoadRecord->RecordAddress	= dwLinAddr + dwSegAddr;
      	pLoadRecord->RecordLength	= RecordLen + sizeof(HEX_HEADER_T);

	Socketioc.length = pLoadRecord->RecordLength;
	Socketioc.arg = pLoadRecord;

      	NTStatus = SendIoctlCmd(SocketID, TIG_LOAD_RECORD, &Socketioc);
      	break;

      case EXT_SEG_ADDR_REC:
        //*
        //* Extended Segment Address
        //*
        dwSegAddr = atoh(&pBuffer[sizeof(HEX_HEADER_T)], RecordLen) << 4;
        break;

      case EXT_LIN_ADDR_REC:
        //*
        //* Extended Linear Address
        //*
        dwLinAddr = atoh(&pBuffer[sizeof(HEX_HEADER_T)], RecordLen) << 16;
        break;

      case EXT_GOTO_CMD_REC:
        //*
        //* Start Segment Address (GOTO)
        //*
        dwIPValue = atoh(&pBuffer[sizeof(HEX_HEADER_T)], RecordLen);

        pLoadRecord->RecordType    = EXT_GOTO_CMD_REC;
        pLoadRecord->RecordAddress = dwIPValue;
        pLoadRecord->RecordLength  = 0;

	Socketioc.length = sizeof(TIG_LOAD_RECORD_T);
	Socketioc.arg = pLoadRecord;

        // execute the record
        NTStatus = SendIoctlCmd(SocketID, TIG_LOAD_RECORD, &Socketioc);
        break;

      case EXT_END_OF_FILE:
        //*
        //* End Of File
        //*
        FoundLastRecord = TRUE;
        break;

      default:
        //*
        //* Unknown Record
        //*
        NTStatus = STATUS_FAILURE;
        printf("Error - Invalid record type, Load terminating\n");
        break;
    }
  } while (!FoundLastRecord && NT_SUCCESS(NTStatus));

  if (pLoadRecord)
    free(pLoadRecord);

  //*
  //* Make Sure Download completed
  //*
  if (!FoundLastRecord)
  {
    printf("Last record not found - load terminating\n");
    NTStatus = STATUS_FAILURE;
  }

  return NTStatus;
}

/******************************************************************************
FUNCTION NAME:
	atoh

ABSTRACT:
	Converts ASCII hex to DWORD

PARAMETERS:
	char *cInStr = pointer to string
	DWORD uNumChars = number of characters in the string

RETURN:
	none

DETAILS:
******************************************************************************/
DWORD
atoh(
	const char *	cInStr,
	DWORD		uNumChars)
{
  DWORD uSum = 0;

  while (uNumChars--) {
    char ch = *cInStr++;

    if (ch >= '0' &&
	 ch <= '9')
      ch -= '0';
    else if (ch >= 'A' &&
	      ch <= 'F')
      ch -= ('A' - 10);
    else if (ch >= 'a' &&
	      ch <= 'f')
      ch -= ('a' - 10);
    else
      continue;

    uSum <<= 4;
    uSum += ch;
  }

  return uSum;
}

/******************************************************************************
FUNCTION NAME:
	SendIoctlCmd

ABSTRACT:
	This function is called to send a load record to the
	modem device.

PARAMETERS:
	SocketID - identifies the socket to send to
	CommandID - ?
	Socketioc->arg - pointer to the record data to send
	Socketioc->length - length of the record in bytes

RETURN:
	The status of the operation.

DETAILS:
******************************************************************************/
NTSTATUS
SendIoctlCmd(
	IN int		SocketID,
	IN int		CommandID,
	struct atmif_sioc	*pSocketioc)
{
  int				Status;

  if (d_opt) {
    printf("About to ioctl %x=>%i,%i,%p\n",
	   CommandID,
	   pSocketioc->number, pSocketioc->length, pSocketioc->arg);
    fflush(NULL);
  }

  Status = ioctl(SocketID, CommandID, pSocketioc);
  if (Status < 0)
  {
    printf(" bad ioctl(%x, %x,), Status=%x\n", SocketID, CommandID, Status);
    return STATUS_FAILURE;
  } else
    return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	SendUserParams

ABSTRACT:
	This function is called to send a user configurable parameters
	to the device driver.

PARAMETERS:
	SocketID - identifies the socket to send to

RETURN:
	The status of the operation.

DETAILS:
******************************************************************************/
NTSTATUS
SendUserParams(
	IN  int				SocketID,
	IN  const char *		Path,
	OUT unsigned long *		pAutoLogEnable,
	OUT char *			AutoLogPath,
	OUT char *			FirmwareFilename,
	OUT unsigned long *		pAutoLogFrequency,
	OUT unsigned long *		pAutoLogMaxSize,
	OUT unsigned long *		pAutoLogOverwrite)
{
  TIG_USER_PARAMS	TigParams;
  NTSTATUS		NTStatus;

  NTStatus = ReadUserParamsE(Path,
			     pAutoLogEnable,
			     AutoLogPath,
			     FirmwareFilename,
			     pAutoLogFrequency,
			     pAutoLogMaxSize,
			     pAutoLogOverwrite,
			     &TigParams);

  if (NTStatus != STATUS_SUCCESS)
    return STATUS_FAILURE;

  Socketioc.length = sizeof(TigParams);
  Socketioc.arg = &TigParams;

  // send the data to the driver
  NTStatus = SendIoctlCmd(SocketID, TIG_SET_PARAMS, &Socketioc);
  if (NT_SUCCESS(NTStatus) &&
      !NT_SUCCESS(TigParams.CommandStatus))
  {
    NTStatus = STATUS_FAILURE;
    printf("TIG_SET_PARAMS IOCTL returned error in parameters, CommandStatus=%lx\n", TigParams.CommandStatus);
  }

  return NTStatus;
}
